from __future__ import annotations
from typing import Optional

from argparse import ArgumentParser
from dataclasses import asdict
from pprint import pprint

from .config import compute_operations, Config
from .sync_plan import SyncError
from .version import __version__


class App:
    @classmethod
    def run(cls, parser: ArgumentParser):
        cmd = parser.add_subparsers(title="command", dest="cmd")
        parser.set_defaults(run=App.print_help(parser))

        ##################################################
        # Version
        ##################################################

        version_cmd = cmd.add_parser(
            "version", help="print version and exit",
        )
        version_cmd.set_defaults(run=App.run_version)

        ##################################################
        # Link
        ##################################################

        link_cmd = cmd.add_parser(
            "link", help=f"link dotfiles from {Config.root()}",
            description=f"links files into the home directory."
        )
        link_cmd.set_defaults(run=App.run_link)

        link_cmd.add_argument(
            '-c', dest='category', default='common',
            help='specify a category to sync (defaults to common)'
        )
        link_cmd.add_argument(
            '-t', dest='topic', default=None,
             help='specify a topic to sync (inside a category)'
        )
        link_cmd.add_argument(
            '-f', dest='force', action='store_true',
             help='Force execution'
        )
        link_cmd.add_argument(
            '-d', dest='dry_run', action='store_true',
            help='dry run current setup'
        )
        link_cmd.add_argument(
            '-b', dest='backup', action='store_false', default=True,
             help='backup files and place new ones in place, appends ".backup"'
        )

        ##################################################
        # Config
        ##################################################

        config_cmd = cmd.add_parser(
            "config", help="query configuration values",
            description="Query configuration values and sync plan."
        )
        config_cmd.set_defaults(run=App.print_help(config_cmd))

        config_subcommand = config_cmd.add_subparsers(title="command", dest="subcmd")

        ##################################################
        # Config/Root
        ##################################################

        config_root_cmd = config_subcommand.add_parser(
            "root", help="show root location"
        )
        config_root_cmd.set_defaults(run=App.run_config_root)

        ##################################################
        # Config/List
        ##################################################

        config_list_cmd = config_subcommand.add_parser(
            "list", help="list files and sync plan withing a category/topic"
        )
        config_list_cmd.set_defaults(run=App.run_config_list)

        config_list_cmd.add_argument(
            "--detailed", action="store_true",
            help="print detailed sync plan"
        )

        config_list_cmd.add_argument("path", metavar="PATHS", nargs="?")

        ##################################################
        # Config/Dump
        ##################################################

        config_dump_cmd = config_subcommand.add_parser(
            "dump", help="dump parsed configuration for a particular category"
        )
        config_dump_cmd.set_defaults(run=App.run_config_dump)

        config_dump_cmd.add_argument("category", metavar="CATEGORY")

        ns = parser.parse_args()
        return ns.run(**vars(ns))

    @staticmethod
    def print_help(parser):
        return lambda **kwargs: parser.print_help()

    @staticmethod
    def run_version(**kwargs):
        print(__version__)

    @staticmethod
    def run_link(category: str, topic: str, dry_run: bool, backup: bool, force: bool, **kwargs):
        conf = Config.load(category)
        if conf is None:
            return
        for op_topic, ops in compute_operations(conf).items():
            if topic is not None and op_topic != topic:
                continue
            print(f":: Linking({op_topic})")
            for op in ops:
                for plan in op.reconcile():
                    if dry_run:
                        plan.print()
                    else:
                        plan.apply(force=force, backup=backup)

    @staticmethod
    def run_config_root(**kwargs):
        print(Config.root())
        return

    @staticmethod
    def run_config_list(path: Optional[str], detailed: bool, **kwargs):
        if path is None:
            for category in Config.categories():
                print(f"Category({category}):")
                for topic in Config.load(category).config.topics:
                    print(f"  Topic({topic})")
            return

        components = path.split("/")
        while len(components) < 2:
            components.append("")

        category, topic, *suffix = components
        conf = Config.load(category)

        for op_topic, ops in compute_operations(conf).items():
            if topic != "" and op_topic != topic:
                continue
            print(f"Topic({op_topic}):")
            for op in ops:
                prefix_path = conf.config.root.joinpath(topic, *suffix)
                if op.src_path.is_relative_to(prefix_path):
                    print(f"  Path({op.src_path})")
                    print(f"    {op}")

                if detailed:
                    for plan in op.reconcile():
                        print(f"      {plan}")

    @staticmethod
    def run_config_dump(category: str, **kwargs):
        conf = Config.load(category)
        pprint(asdict(conf), indent=2)


from __future__ import annotations, absolute_import
from typing import Union, Dict, List, Optional, Set
from enum import Enum
from dataclasses import dataclass

import json
import os

from pathlib import PosixPath

# import dacite
# from dacite import from_dict

from .sync_plan import LogicalSyncPlan
from .utils import coalesce, path_matches_patterns, typecast, typecast_json_file

CONFIG_NAME = "dot.conf.json"
CONFIG_PATH = os.getenv("DOTTER_CONFIG_ROOT", "~/.config/dotter")

CONFIG_STANDARD = {
    "defaults": {
        "root": "~/",

        "add_dot": True,
        "use_contents": True,

        "link_mode": "recursive_link",
        "recursive_modifiers": {
            "link": ".xlink",
            "touch": ".xtouch",
            "copy": ".xcopy"
        },
        "ignore": [
            "*/.*",
            "*/.*/*",
        ]
    },
}

CONFIG_STANDARD_DEFAULTS = CONFIG_STANDARD.get("defaults")

class ConfigLinkMode(Enum):
    RLINK = "recursive_link"
    LINK = "link"
    RCOPY = "recursive_copy"
    COPY = "copy"
    TOUCH = "touch"

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f"M({self.value})"

    def is_recursive(self):
        if self in [self.RLINK, self.RCOPY]:
            return True
        return False

    def unrecurse(self):
        return {
            self.RCOPY: self.COPY,
            self.RLINK: self.LINK,
        }.get(self, self)


@dataclass
class ConfigPatternSetting:
    root: Union[str, PosixPath] = None
    add_dot: bool = None
    use_contents: bool = None
    link_mode: ConfigLinkMode = None
    ignore: List[str] = None
    recursive_modifiers: Dict[ConfigLinkMode, str] = None

    def __post_init__(self):
        if self.root is not None:
            self.root = PosixPath(self.root).expanduser().resolve()

    def merge(self, override) -> ConfigPatternSetting:
        if override is None:
            return self
        return ConfigPatternSetting(
            root=coalesce(override.root, self.root),
            add_dot=coalesce(override.add_dot, self.add_dot),
            use_contents=coalesce(override.use_contents, self.use_contents),
            link_mode=coalesce(override.link_mode, self.link_mode),
            ignore=coalesce(override.ignore, self.ignore),
            recursive_modifiers=coalesce(override.recursive_modifiers, self.recursive_modifiers),
        )


@dataclass
class ConfigCategory:
    defaults: ConfigPatternSetting = None
    topics: Dict[str, ConfigPatternSetting] = None
    disabled: List[str] = None

    _root: PosixPath = None

    @property
    def root(self) -> PosixPath:
        return self._root

    @root.setter
    def root(self, path: PosixPath):
        self._root = path

    def __post_init__(self):
        if self.defaults is None:
            self.defaults = ConfigPatternSetting()
        if self.topics is None:
            self.topics = {}
        if self.disabled is None:
            self.disabled = []

    def merge_defaults(self):
        defaults = typecast(CONFIG_STANDARD_DEFAULTS, ConfigPatternSetting)
        self.defaults = defaults.merge(self.defaults)

    def merge_topics(self):
        for topic, override in self.topics.items():
            self.topics[topic] = self.defaults.merge(override)


@dataclass
class Config:
    config: ConfigCategory

    @staticmethod
    def root():
        return PosixPath(os.path.normpath(CONFIG_PATH)).expanduser()

    @staticmethod
    def from_dict(d: dict):
        return typecast(d, ConfigCategory)

    @staticmethod
    def categories():
        config_root = Config.root()
        return [
            p.relative_to(config_root)
            for p in config_root.iterdir()
            if p.is_dir() and not p.name.startswith(".")
        ]

    @staticmethod
    def load(category: str = None) -> Optional[Config]:
        config_root_path = Config.root()
        config_category_path = config_root_path.joinpath(category)
        config_file_path = config_category_path.joinpath(CONFIG_NAME)

        config_data = typecast_json_file(config_file_path, ConfigCategory, ConfigCategory())

        config_data.root = config_category_path
        config_data.merge_defaults()

        for contents in config_category_path.iterdir():
            if not contents.is_dir():
                continue
            if contents.name.startswith("."):
                continue
            if contents.name not in config_data.topics:
                config_data.topics[contents.name] = ConfigPatternSetting()

        config_data.merge_topics()

        conf = Config(
            config=config_data,
        )

        return conf


def compute_operations(category: Config) -> Dict[str, List[LogicalSyncPlan]]:
    ops: Dict[str, List[LogicalSyncPlan]] = {}
    for topic, topic_config in category.config.topics.items():
        config_topic_path = category.config.root.joinpath(topic)

        if not config_topic_path.exists():
            continue

        # Do we use contents of the folder?
        if topic_config.use_contents:
            # If we do then loop over all dirs
            link_items = list(config_topic_path.iterdir())
        else:
            # If we dont then only visit the toplevel dir.
            link_items = [config_topic_path]

        ops[topic] = compute_topic_operations(link_items, topic_config)
    return ops


def compute_topic_operations(link_items: List[PosixPath], link_config: ConfigPatternSetting):
    topic_ops: List[LogicalSyncPlan] = []

    for link_item in link_items:
        op_type, src_path, dst_path = _determine_operation(
            link_config, link_item.parent, link_item,
        )

        # Recursive link/copy mode is the most involved
        # - Find all recursive link modifiers in paths under link_item and adjust their link_mode.
        if op_type.is_recursive():
            # Break up paths by simple operations:
            link_paths = link_item.glob("**/*")
            link_paths = filter(lambda p: not path_matches_patterns(p, link_config.ignore), link_paths)

            seen_prefixes: Set[str] = set()
            for link_path in link_paths:
                op_type_switch, src_path, dst_path = _determine_operation(
                    link_config, link_item.parent, link_path,
                )

                if str(src_path) in seen_prefixes:
                    continue
                if src_path.is_dir() and op_type == op_type_switch:
                    continue

                seen_prefixes.add(str(src_path))

                topic_ops.append(LogicalSyncPlan(
                    type=op_type_switch,
                    src_path=src_path,
                    dst_path=dst_path,
                ))
        else:
            # Simple case, link item is directly linked, copied or touched.
            topic_ops.append(LogicalSyncPlan(
                type=op_type,
                src_path=src_path,
                dst_path=dst_path,
            ))

    return topic_ops


def _determine_operation(link_config: ConfigPatternSetting, link_root_path: PosixPath, link_path: PosixPath):
    op_type, prefix_path, src_path, suffix = _split_by_modifiers(str(link_path), link_config.recursive_modifiers)
    op_type = coalesce(op_type, link_config.link_mode)

    if op_type.is_recursive() and not link_path.is_dir():
        op_type = op_type.unrecurse()

    link_path = PosixPath(src_path)
    link_prefix_path = PosixPath(prefix_path)

    src_path = link_path
    dst_path = _rename_path(link_prefix_path, link_root_path, link_config.root, link_config.add_dot)

    return op_type, src_path, dst_path,


def _rename_path(path: PosixPath, base_path: PosixPath, new_base_path: PosixPath, add_dot: bool):
    rel_path = path.relative_to(base_path)
    if add_dot:
        rel_path = PosixPath("." + str(rel_path))
    return new_base_path.joinpath(rel_path)


def _split_by_modifiers(path: str, modifiers: Dict[ConfigLinkMode, str]) -> (ConfigLinkMode, str, str, str):
    (rext, prefix_path, src_path, suffix) = (None, path, path, "")
    for ext_name, ext in modifiers.items():
        idx = path.find(ext)
        if idx > 0:
            rext = ext_name
            prefix_path = path[0:idx]
            src_path = path[0:idx + len(ext)]
            suffix = path[idx + len(ext):]
    return rext, prefix_path, src_path, suffix



from __future__ import annotations
from typing import Iterable, TypeVar, Dict, Type, Optional
from enum import Enum

import json
from pathlib import PosixPath

from dacite import from_dict, Config


T = TypeVar('T')


def typecast(d: Dict, data_class: Type[T]) -> T:
    return from_dict(data_class=data_class, data=d, config=Config(cast=[Enum]))


def typecast_json_file(path: PosixPath, data_class: Type[T], if_not_found: Optional[T] = None) -> Optional[T]:
    try:
        with open(path) as conf_fd:
            return typecast(json.load(conf_fd), data_class)
    except FileNotFoundError:
        return if_not_found


def coalesce(*args: T) -> T:
    for arg in args:
        if arg is not None:
            return arg


def path_matches_patterns(path: PosixPath, patterns: Iterable[str]) -> bool:
    for ignore_pattern in patterns:
        if path.match(ignore_pattern):
            return True
    return False



from __future__ import annotations
from typing import Dict, List

import shutil
import hashlib

from dataclasses import dataclass
from pathlib import PosixPath

from . import colors

class SyncError(RuntimeError):
    pass

@dataclass
class PhysicalSyncPlan:
    type: str
    action: str
    src_path: PosixPath
    dst_path: PosixPath

    def __str__(self):
        if self.type == "link":
            return f"PLAN({self.type}.{self.action} {str(self.src_path)} <- {self.dst_path})"
        else:
            return f"PLAN({self.type}.{self.action} {str(self.src_path)} -> {self.dst_path})"

    def print(self):
        if self.action == "create":
            print(colors.yellow(str(self)))
        elif self.action == "replace":
            print(colors.red(str(self)))

    def log(self, skip: bool = False):
        if skip:
            print(colors.blue("SKIPPING", str(self), "(force to apply)"))
        else:
            print(colors.yellow("APPLYING", str(self)))

    def apply(self, force: bool = False, backup: bool = True):
        if self.type == "dir":
            return self.apply_dir(force, backup)
        elif self.type == "touch":
            return self.apply_touch(force, backup)
        elif self.type == "link":
            return self.apply_link(force, backup)
        elif self.type == "copy":
            return self.apply_copy(force, backup)

    def apply_touch(self, force: bool, backup: bool):
        if self.src_path.is_file():
            self.log()
            self._copy_path(force=force, backup=backup)

    def apply_dir(self, force: bool, backup: bool):
        if self.action == "create":
            self.log()
            self.dst_path.mkdir(exist_ok=True)
        elif self.action == "replace":
            self.log(skip=not force)
            if force:
                self._backup_dest_path(backup)
                self.dst_path.mkdir(exist_ok=True)

    def apply_link(self, force: bool, backup: bool):
        if self.action == "create":
            self.log()
            self.dst_path.symlink_to(self.src_path)
        elif self.action == "replace":
            self.log(skip=not force)
            if force:
                self._backup_dest_path(backup)
                self.dst_path.symlink_to(self.src_path)

    def apply_copy(self, force: bool, backup: bool):
        if self.action == "create":
            self.log()
            self._copy_path(force=force, backup=backup)
        elif self.action == "replace":
            self.log(skip=not force)
            if force:
                self._backup_dest_path(backup)
                self._copy_path(force=force, backup=backup)

    def _backup_dest_path(self, backup: bool):
        if backup:
            dst_path_bak = PosixPath(str(self.dst_path) + ".bak")
            if dst_path_bak.exists():
                raise SyncError(f"file {dst_path_bak} already exists, clean up to continue")
            shutil.move(self.dst_path, dst_path_bak)
        else:
            self.dst_path.unlink(missing_ok=True)

    def _copy_path(self, force: bool, backup: bool):
        if self.src_path.is_file():
            shutil.copy2(self.src_path, self.dst_path)
        else:
            shutil.copytree(self.src_path, self.dst_path)


@dataclass
class LogicalSyncPlan:
    type: str
    src_path: PosixPath
    dst_path: PosixPath
    # debug: str = None

    def reconcile(self) -> List[PhysicalSyncPlan]:
        if self.type == "touch":
            return self.reconcile_touch()
        elif self.type == "link":
            return self.reconcile_link()
        elif self.type == "copy":
            return self.reconcile_copy()
        return []

    def reconcile_dir(self, dst: PosixPath) -> List[PhysicalSyncPlan]:
        ops = []
        for parent in reversed(dst.parents):
            if not parent.exists():
                ops.append(PhysicalSyncPlan(
                    type="dir", action="create",
                    dst_path=parent,
                    src_path=parent
                ))
            elif not parent.is_dir():
                ops.append(PhysicalSyncPlan(
                    type="dir", action="replace",
                    dst_path=parent,
                    src_path=parent
                ))
        return ops

    def reconcile_touch(self) -> List[PhysicalSyncPlan]:
        ops = []
        ops.extend(self.reconcile_dir(self.dst_path))

        if self.dst_path.exists():
            return []

        ops.append(PhysicalSyncPlan(
            type="touch", action="create",
            src_path=self.src_path,
            dst_path=self.dst_path,
        ))
        return ops

    def reconcile_link(self) -> List[PhysicalSyncPlan]:
        ops = []
        ops.extend(self.reconcile_dir(self.dst_path))

        if not self.dst_path.exists():
            ops.append(PhysicalSyncPlan(
                type="link", action="create",
                src_path=self.src_path,
                dst_path=self.dst_path,
            ))
        elif not _check_link_points_to(self.dst_path, self.src_path):
            ops.append(PhysicalSyncPlan(
                type="link", action="replace",
                src_path=self.src_path,
                dst_path=self.dst_path,
            ))
        return ops

    def reconcile_copy(self):
        ops = []
        ops.extend(self.reconcile_dir(self.dst_path))

        if not self.dst_path.exists():
            ops.append(PhysicalSyncPlan(
                type="copy", action="create",
                src_path=self.src_path,
                dst_path=self.dst_path,
            ))
        elif not _check_paths_same_type(self.src_path, self.dst_path):
            ops.append(PhysicalSyncPlan(
                type="copy", action="replace",
                src_path=self.src_path,
                dst_path=self.dst_path,
            ))
        elif self.src_path.is_file():
            ops.append(PhysicalSyncPlan(
                type="copy", action="replace",
                src_path=self.src_path,
                dst_path=self.dst_path,
            ))
        elif self.src_path.is_dir():
            missing_files, existing_files = _get_missing_files(self.src_path, self.dst_path)
            for f in missing_files:
                dst_path = self.dst_path.joinpath(f)
                src_path = self.src_path.joinpath(f)
                ops.extend(self.reconcile_dir(dst_path))
                ops.append(PhysicalSyncPlan(
                    type="copy", action="create",
                    src_path=src_path,
                    dst_path=dst_path,
                ))
            for f in existing_files:
                dst_path = self.dst_path.joinpath(f)
                src_path = self.src_path.joinpath(f)
                if not _check_file_equal(src_path, dst_path):
                    ops.append(PhysicalSyncPlan(
                        type="copy", action="replace",
                        src_path=src_path,
                        dst_path=dst_path,
                    ))
        return ops


def _check_paths_same_type(a: PosixPath, b: PosixPath) -> bool:
    if a.is_file() and b.is_file():
        return True
    if a.is_dir() and b.is_dir():
        return True
    return False


def _check_link_points_to(src: PosixPath, dst: PosixPath) -> bool:
    if src.is_symlink() and src.resolve() == dst.resolve():
        return True
    return False


def _check_dir_equal(src: PosixPath, dst: PosixPath):
    missing_files, _ = _get_missing_files(src, dst)
    if len(missing_files) > 0:
        return False
    return True


def _check_file_equal(a: PosixPath, b: PosixPath):
    if not a.exists() or not b.exists():
        return False
    if not a.is_file() or not b.is_file():
        return False
    ha = hashlib.md5(open(a, 'rb').read()).hexdigest()
    hb = hashlib.md5(open(b, 'rb').read()).hexdigest()
    return ha == hb


def _get_missing_files(src: PosixPath, dst: PosixPath):
    subpath = lambda base: lambda p: p.relative_to(base)
    is_file = lambda p: p.is_file()

    src_files = set(map(subpath(src), filter(is_file, src.glob("**/*"))))
    dst_files = set(map(subpath(dst), filter(is_file, dst.glob("**/*"))))

    existing = dst_files.intersection(src_files)
    missing = src_files.difference(existing)

    return sorted(missing), sorted(existing)



from setuptools import setup, find_packages

__version__ = None

with open('dotter/version.py') as f:
     exec(f.read())

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="nl-dotter",
    version=__version__,
    url="https://github.com/NonLogicalDev/cli.dotter",
    license="MIT",

    author="nonlogicaldev",
    description="A dotfile link farm manager.",

    long_description=long_description,
    long_description_content_type="text/markdown",

    scripts=["bin/dotter"],

    packages=find_packages(),
    install_requires=[
        'dacite',
    ],
    keywords=[
        "dotfile", "link farm"
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
)



HEADER = '\033[95m'

BLUE = '\033[94m'
GREEN = '\033[92m'
YELLOW = '\033[93m'
RED = '\033[91m'

BOLD = '\033[1m'
UNDERLINE = '\033[4m'

ENDC = '\033[0m'


def red(*s):
    return RED + str(" ".join(s)) + ENDC

def green(*s):
    return GREEN + str(" ".join(s)) + ENDC

def yellow(*s):
    return YELLOW + str(" ".join(s)) + ENDC

def blue(*s):
    return BLUE + str(" ".join(s)) + ENDC
